import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, TypeVar, final

from pydantic import BaseModel
from pydantic.color import Color
from websockets.client import WebSocketClientProtocol

logger = logging.getLogger(__file__)


class BaseEntityHandlerConfig(BaseModel):
    entity_id: str


RawMessage = Any

Config = TypeVar("Config", bound=BaseEntityHandlerConfig, covariant=True)
State = TypeVar("State", covariant=True)


class BaseEntityHandler(Generic[Config, State], ABC):
    @final
    def __init__(self, config: Config):
        self.config: Config = config

        self._state: State = self.get_initial_state()

    @staticmethod
    @abstractmethod
    def get_initial_state() -> State:
        """
        Used to initialize local state before any messages are received.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_state_from_message(raw_message: RawMessage) -> State:
        pass

    @abstractmethod
    def on_state_change(self, set_color: Callable[[Color], None]) -> None:
        """
        Performs a side-effect when the state is changed.
        This should generally be used to change the color of the keypad light.
        """
        pass

    @abstractmethod
    async def on_keypress(self, ws: WebSocketClientProtocol, message_id: int) -> None:
        pass

    def get_current_state(self) -> State:
        return self._state

    def on_message(
        self, raw_message: RawMessage, set_color: Callable[[Color], None]
    ) -> None:
        logger.info(f"Entity {self.config.entity_id} updated")
        self._state = self.get_state_from_message(raw_message)
        self.on_state_change(set_color)



import asyncio
import logging
from typing import Any, Callable, Dict, List, Literal, Mapping, Optional, Tuple, Type

import keybow
import websockets
from pydantic import BaseModel, SecretStr, ValidationError
from pydantic.color import Color

from hasspad.config import HasspadConfig
from hasspad.handlers.base import BaseEntityHandler, BaseEntityHandlerConfig

logger = logging.getLogger(__file__)


class AuthRequiredMessage(BaseModel):
    type: Literal["auth_required"]


class AuthTokenMessage(BaseModel):
    type: Literal["auth"] = "auth"
    access_token: SecretStr

    class Config:
        json_encoders: Mapping[
            Type[SecretStr], Callable[[SecretStr], Optional[str]]
        ] = {
            SecretStr: lambda v: v.get_secret_value() if v else None,
        }


class AuthResponseMessage(BaseModel):
    type: Literal["auth_ok", "auth_invalid"]
    message: Optional[str]


class GetStatesMessage(BaseModel):
    id: int
    type: Literal["get_states"] = "get_states"


class GetStatesResponseMessage(BaseModel):
    id: int
    type: Literal["result"]
    success: bool
    result: List[Dict[str, Any]]


class SubscribeEventsMessage(BaseModel):
    id: int
    type: Literal["subscribe_events"] = "subscribe_events"
    event_type: Literal["state_changed"] = "state_changed"


class SubscribeEventsResponseMessage(BaseModel):
    id: int
    type: Literal["result"]
    success: bool


class StateChangeData(BaseModel):
    new_state: Any


class StateChangeEvent(BaseModel):
    event_type: Literal["state_changed"]
    data: StateChangeData


class StateChangeEventMessage(BaseModel):
    type: Literal["event"]
    event: StateChangeEvent


EntityHandler = BaseEntityHandler[BaseEntityHandlerConfig, object]


class Hasspad:
    def __init__(self, config: HasspadConfig):
        self.config = config
        self.idx_to_handler, self.entity_id_to_idx = self._get_entity_handlers()

        self._message_id = 0
        self._message_id_lock = asyncio.Lock()

    async def listen(self):
        try:
            keybow.setup(self.config.size.get_keymap())

            async with websockets.connect(self.config.api.websocket_uri) as ws:
                await self._authenticate(ws)
                await self._initialize_entity_handlers(ws)
                await self._subscribe_to_state_changes(ws)

                keybow.on(handler=self._get_keybow_handler(ws))

                while True:
                    try:
                        message = StateChangeEventMessage.parse_raw(await ws.recv())
                    except ValidationError:
                        continue

                    new_state = message.event.data.new_state
                    self._handle_message(new_state["entity_id"], new_state)

        except Exception:
            keybow.set_all(255, 0, 0)
            keybow.show()
            raise

    async def _get_next_message_id(self):
        async with self._message_id_lock:
            self._message_id += 1
            return self._message_id

    def _get_entity_handlers(
        self,
    ) -> Tuple[Mapping[int, EntityHandler], Mapping[str, int],]:
        idx_to_handler = {}
        entity_id_to_idx = {}

        for idx, handler_config in enumerate(self.config.handlers):
            if handler_config is None:
                continue

            handler = handler_config.get_handler()

            idx_to_handler[idx] = handler
            entity_id_to_idx[handler_config.entity_id] = idx

        return idx_to_handler, entity_id_to_idx

    async def _authenticate(
        self, ws: websockets.client.WebSocketClientProtocol
    ) -> None:
        # 2.1: Server sends auth_required

        AuthRequiredMessage.parse_raw(await ws.recv())

        # 2.2: Client sends auth

        await ws.send(
            AuthTokenMessage(access_token=self.config.api.access_token).json()
        )

        # 2.3: Server sends auth_ok or auth_invalid

        auth_response_msg = AuthResponseMessage.parse_raw(await ws.recv())
        if auth_response_msg.type == "auth_invalid":
            raise Exception(auth_response_msg.message)

        logger.info("Authentication successful")

    async def _initialize_entity_handlers(
        self, ws: websockets.client.WebSocketClientProtocol
    ) -> None:
        request = GetStatesMessage(
            id=await self._get_next_message_id(),
        )
        await ws.send(request.json())

        response = GetStatesResponseMessage.parse_raw(await ws.recv())

        assert response.success == True

        for result in response.result:
            self._handle_message(result["entity_id"], result)

    async def _subscribe_to_state_changes(
        self, ws: websockets.client.WebSocketClientProtocol
    ) -> None:
        request = SubscribeEventsMessage(id=await self._get_next_message_id())
        await ws.send(request.json())

        response = SubscribeEventsResponseMessage.parse_raw(await ws.recv())

        assert response.success == True

    def _get_keybow_handler(
        self, ws: websockets.client.WebSocketClientProtocol
    ) -> Callable[[int, bool], None]:
        async def on_keypress(idx: int):
            entity_handler = self.idx_to_handler.get(idx)
            if not entity_handler:
                return

            message_id = await self._get_next_message_id()
            await entity_handler.on_keypress(ws, message_id)

        def handler(idx: int, key_state: bool):
            if key_state:
                return

            asyncio.run(on_keypress(idx))

        return handler

    def _set_key_color(self, idx: int, color: Color) -> None:
        logger.info(f"Setting key {idx} to {color}")
        keybow.set_led(idx, *color.as_rgb_tuple(alpha=False))
        keybow.show()

    def _handle_message(self, entity_id: str, message: Any) -> None:
        handler_idx = self.entity_id_to_idx.get(entity_id)
        if handler_idx is None:
            return None

        entity_handler = self.idx_to_handler[handler_idx]
        entity_handler.on_message(
            message, lambda color: self._set_key_color(handler_idx, color)
        )



from typing import Callable, Literal, Optional, Tuple, final

from pydantic import BaseModel
from pydantic.color import Color
from websockets.client import WebSocketClientProtocol

from .base import BaseEntityHandler, BaseEntityHandlerConfig, RawMessage


class LightEntityHandlerConfig(BaseEntityHandlerConfig):
    type: Literal["light"]

    def get_handler(self):
        return LightEntityHandler(self)


class LightEntityHandlerState(BaseModel):
    is_on: bool
    color: Color


class LightAttributes(BaseModel):
    rgb_color: Optional[Tuple[int, int, int]]


class LightEntityMessage(BaseModel):
    state: Literal["on", "off"]
    attributes: LightAttributes


class UpdateLightStateMessageServiceData(BaseModel):
    entity_id: str


class UpdateLightStateMessage(BaseModel):
    id: int
    service: Literal["turn_off", "turn_on"]
    service_data: UpdateLightStateMessageServiceData

    type = "call_service"
    domain = "light"


@final
class LightEntityHandler(
    BaseEntityHandler[LightEntityHandlerConfig, LightEntityHandlerState]
):
    @staticmethod
    def get_initial_state() -> LightEntityHandlerState:
        return LightEntityHandlerState(is_on=False, color=(0, 0, 0))

    @staticmethod
    def get_state_from_message(raw_message: RawMessage) -> LightEntityHandlerState:
        message = LightEntityMessage(**raw_message)
        return LightEntityHandlerState(
            is_on=message.state == "on",
            color=message.attributes.rgb_color or (255, 255, 255),
        )

    def on_state_change(self, set_color: Callable[[Color], None]) -> None:
        state = self.get_current_state()
        set_color(state.color if state.is_on else Color((0, 0, 0)))

    async def on_keypress(self, ws: WebSocketClientProtocol, message_id: int) -> None:
        state = self.get_current_state()
        message = UpdateLightStateMessage(
            id=message_id,
            service=("turn_off" if state.is_on else "turn_on"),
            service_data=UpdateLightStateMessageServiceData(
                entity_id=self.config.entity_id
            ),
        )
        await ws.send(message.json())



#!/usr/bin/env python3

import asyncio
import logging
from typing import TextIO

import click
import yaml
from rich.logging import RichHandler

from hasspad.config import HasspadConfig
from hasspad.hasspad import Hasspad

logging.basicConfig(
    level="INFO",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)],
)

logger = logging.getLogger(__file__)


@click.command()
@click.argument("config", type=click.File("r"))
def main(config: TextIO) -> None:
    hasspad = Hasspad(HasspadConfig(**yaml.safe_load(config)))
    asyncio.run(hasspad.listen())


if __name__ == "__main__":
    main(auto_envvar_prefix="HASSPAD")



from enum import Enum
from typing import List, Tuple, Union, cast

import keybow
from pydantic import BaseModel, SecretStr, stricturl, validator

from hasspad.handlers.light import LightEntityHandlerConfig


class APIConfig(BaseModel):
    websocket_uri: stricturl(allowed_schemes={"ws", "wss"})
    access_token: SecretStr


class KeybowSize(Enum):
    MINI = "mini"
    FULL = "full"

    def get_keymap(self) -> List[Tuple[int, int]]:
        if self == KeybowSize.MINI:
            return keybow.MINI
        else:
            return keybow.FULL

    def num_keys(self) -> int:
        return len(self.get_keymap())


HasspadHandlerConfig = Union[LightEntityHandlerConfig]


class HasspadConfig(BaseModel):
    api: APIConfig
    size: KeybowSize
    handlers: List[Union[None, HasspadHandlerConfig]]

    @validator("handlers")
    def handler_count_must_match_size(
        cls, handlers: List[HasspadHandlerConfig], values
    ):
        size = cast(KeybowSize, values["size"])

        if size.num_keys() != len(handlers):
            raise ValueError(
                f"{size} should have {size.num_keys()} handlers, {len(handlers)} provided"
            )

        return handlers



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hasspad', 'hasspad.handlers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'RPi.GPIO>=0.7.0,<0.8.0',
 'click>=7.1.2,<8.0.0',
 'keybow>=0.0.3,<0.0.4',
 'pydantic>=1.7.3,<2.0.0',
 'rich>=9.8.2,<10.0.0',
 'spidev>=3.5,<4.0',
 'websockets>=8.1,<9.0']

entry_points = \
{'console_scripts': ['hasspad = hasspad.main:main']}

setup_kwargs = {
    'name': 'hasspad',
    'version': '0.1.3',
    'description': '',
    'long_description': '# hasspad\n\nA macropad interface for Home Assistant using the Pimoroni Keybow\n\n## Installation\n\nFirst, provision the device. Flash `raspios` and ensure `ssh` and WiFi access are enabled.\nThis can be done by writing a blank `ssh` file and a `wpa_supplicant.conf` file to the BOOT mount.\n\nOnce you are able to login to the device, perform some basic setup to set the hostname and enable SPI:\n\n```\nsudo raspi-config nonint do_hostname hasspad\nsudo raspi-config nonint do_spi 1\nsudo reboot\n```\n\nInstall `docker` and `docker-compose`:\n\n```\ncurl -sSL https://get.docker.com | sh\nsudo usermod -aG docker pi\n\nsudo apt-get install -y libffi-dev libssl-dev python3 python3-pip\nsudo pip3 install docker-compose\n```\n\nThe last thing you need to do is ensure environment variables are set that are used to authenticate with Home Assistant.\nYou can also pass these in as flags or modify `docker-compose.yml`:\n\n```\nexport HASSPAD_URI="ws://homeassistant.local:8123/api/websocket"\nexport HASSPAD_ACCESS_TOKEN="abcd"\n```\n\nFinally, just run `docker-compose up`.\n',
    'author': 'Kevin Chen',
    'author_email': 'keffcat@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kvchen/hasspad/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
